/* $Id: e2_toolbar.h 1719 2006-03-02 06:58:15Z tpgww $
 * 
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef E2_TOOLBAR_H
#define E2_TOOLBAR_H

#include "emelfm2.h"
#include "e2_action.h"
#include "e2_option.h"
#include "e2_fileview.h"

#define FOLDBARS

typedef enum
{	//these are ordered for decreasing visual impact when changes
	//are made
	//if order is changed, correspondingly change order of array rows
	//in e2_toolbar_options_register()
	E2_BAR_TASK,
	E2_BAR_PANE1,
	E2_BAR_PANE2,
	E2_BAR_COMMAND,
	E2_BAR_COUNT	//this is the no. of bars, not a 'type'
} E2_BarType;

typedef struct _E2_ToolbarRuntime
{
	gchar *name;	//'internal' name of bar, for referencing (constant or allocated)
	gchar *public_name;	//translated form of name (constant)
	GtkToolbar *toolbar;
	GtkWidget *toolbar_container;
	GtkWidget *toolbar_container_box;
#ifdef FOLDBARS
	GtkWidget *toolbarvbox;	//hbox holding just a dir-line
#endif
	gboolean blocked;  //TRUE when bar's context menu is being created
	gboolean hidden;
	E2_OptionSet *set;
	E2_OptionSet *show;
	E2_OptionSet *tooltips;
	E2_OptionSet *space;
	E2_OptionSet *hori;
	E2_OptionSet *type;
	E2_OptionSet *priority;
	E2_OptionSet *relief;
	E2_OptionSet *same;
	E2_OptionSet *style;
	E2_OptionSet *isize;
	GtkIconSize icon_size;	//working copy of size enumerator (supports 'default')
	GtkAllocation *alloc;
	GtkRequisition req;
	gboolean has_toggle;
	gboolean has_bookmarks;
	gboolean has_command_line;
	gboolean restbtn_shown;
	GtkWidget *button_rest;
#ifdef FOLDBARS
	gint restbtn_width;	//size to use if folded bars in force
	GtkToolItem *dirline_tool;	//tool item for the movable dirline widget
	gint dirline_index;	//index of dirline_tool in the toolbar
	gboolean folded;	//TRUE when dirline_tool is in separate box
//	gint threshold;	//pixel size when we swap to/from separate box
#endif
	GList *bar_items_list;  //list of toolbar buttons (not spacers etc), including hidden toggles
	GtkWidget *menu_starter;	//bar item that is first in the overflow menu
	gboolean reversed;	//TRUE if the config-order of bar items is reversed onscreen
} E2_ToolbarRuntime;

typedef struct _E2_ToolbarData
{
	E2_BarType type;
	gchar *name;	//internal name, constant, not translated
	gchar *public_name;	//translated name, constant
	E2_ToolbarRuntime *rt;	//data struct for the bar
} E2_ToolbarData;

// this is 'extended' data for toggle actions, stored at action.data2
typedef struct _E2_ToggleData
{
	E2_ToolbarRuntime *bar_rt;  //ptr to data for the bar that includes this (instance of a toggle
	E2_TOGGLE_STATUS state;  //signal for how to handle (re)builds 
	gboolean current_state;
	gint button_style;
	GtkWidget *button_box;
	GtkWidget *true_image;
	GtkWidget *false_image;
	GtkWidget *true_label;
	GtkWidget *false_label;
	GtkTooltipsData *tips;
} E2_ToggleData;

//#ifdef GTK26
void e2_toolbar_initialise_space_handler (E2_ToolbarRuntime *rt);
void e2_toolbar_initialise_tearoffs (void);
//#endif
void e2_toobar_menu_position_set (GtkWidget *button, GtkWidget *menu, 
	GtkAllocation *menu_position, E2_ToolbarRuntime *rt);
void e2_toolbar_menu_position (GtkMenu *menu, gint *x, gint *y,
	gboolean *push_in, GtkAllocation *menu_position);
gboolean e2_toolbar_button_toggle (gchar *action_name);
gboolean e2_toolbar_toggle_button_get_state (gchar *action_name);
void e2_toolbar_toggle_button_set_state (gchar *action_name, gboolean state);
void e2_toolbar_toggle_buttons_set_destroyed (E2_ToolbarRuntime *rt);
void e2_toolbar_toggle_filter_button (FileView *view);
void e2_toolbar_initialise (E2_BarType barnum);
void e2_toolbar_create (E2_ToolbarRuntime *rt);
void e2_toolbar_recreate (E2_ToolbarRuntime *rt);
void e2_toolbar_recreate_all (void);
void e2_toolbar_destroy (E2_ToolbarRuntime *rt);

void e2_toolbar_data_create (void);
void e2_toolbar_actions_register (void);
void e2_toolbar_options_register (E2_BarType barnum);
void e2_toolbar_toolbaroptions_register (void);
void e2_toolbar_commandbaroptions_register (void);

#endif //ndef E2_TOOLBAR_H
